"""Tests for `pyshotgrid` default entities."""
import pyshotgrid.sg_default_entities as sde

from .testbase import BaseShotGridTest


class TestSGProject(BaseShotGridTest):
    """Tests for `SGProject` class."""

    def test_shots(self):
        sg_project = sde.SGProject(self.sg, "Project", 1)

        result = sg_project.shots()

        self.assertEqual(4, len(result))
        self.assertTrue(isinstance(result[0], sde.SGShot))

    def test_shots_glob(self):
        sg_project = sde.SGProject(self.sg, "Project", 1)

        result = sg_project.shots("sq111_*")

        self.assertEqual(2, len(result))
        self.assertEqual("sq111_sh1111", result[0]["code"].get())
        self.assertEqual("sq111_sh2222", result[1]["code"].get())

    def test_assets(self):
        sg_project = sde.SGProject(self.sg, "Project", 1)

        result = sg_project.assets()

        self.assertEqual(3, len(result))
        self.assertTrue(isinstance(result[0], sde.SGAsset))

    def test_assets_glob(self):
        sg_project = sde.SGProject(self.sg, "Project", 1)

        result = sg_project.assets("Car*")

        self.assertEqual(2, len(result))
        self.assertEqual("CarA", result[0]["code"].get())
        self.assertEqual("CarB", result[1]["code"].get())

    def test_playlists(self):
        sg_project = sde.SGProject(self.sg, "Project", 1)

        result = sg_project.playlists()

        self.assertEqual(1, len(result))
        self.assertTrue(isinstance(result[0], sde.SGPlaylist))


class TestSGPublishedFile(BaseShotGridTest):
    """Tests for `SGPublishedFile` class."""

    def test_is_latest__false(self):
        sg_publish = sde.SGPublishedFile(self.sg, "PublishedFile", 1)

        result = sg_publish.is_latest()

        self.assertFalse(result)

    def test_is_latest__true(self):
        sg_publish = sde.SGPublishedFile(self.sg, "PublishedFile", 5)

        result = sg_publish.is_latest()

        self.assertTrue(result)

    def test_get_latest_publish__is_already_latest(self):
        sg_publish = sde.SGPublishedFile(self.sg, "PublishedFile", 5)

        result = sg_publish.get_latest_publish()

        self.assertEqual(sg_publish, result)

    def test_get_latest_publish(self):
        sg_publish = sde.SGPublishedFile(self.sg, "PublishedFile", 1)
        latest_sg_publish = sde.SGPublishedFile(self.sg, "PublishedFile", 5)

        result = sg_publish.get_latest_publish()

        self.assertEqual(latest_sg_publish, result)

    def test_get_all_publishes(self):
        sg_publishes = [
            sde.SGPublishedFile(self.sg, "PublishedFile", i) for i in range(1, 6)
        ]

        result = sg_publishes[0].get_all_publishes()

        self.assertEqual(sg_publishes, result)

    def test_get_next_publishes(self):
        sg_publish = sde.SGPublishedFile(self.sg, "PublishedFile", 3)
        result_sg_publishes = [
            sde.SGPublishedFile(self.sg, "PublishedFile", i) for i in range(4, 6)
        ]

        result = sg_publish.get_next_publishes()

        self.assertEqual(result_sg_publishes, result)

    def test_get_previous_publishes(self):
        sg_publish = sde.SGPublishedFile(self.sg, "PublishedFile", 3)
        result_sg_publishes = [
            sde.SGPublishedFile(self.sg, "PublishedFile", i) for i in range(1, 3)
        ]

        result = sg_publish.get_previous_publishes()

        self.assertEqual(result_sg_publishes, result)



"""Tests for `pyshotgrid` SGSite class."""
from shotgun_api3.lib import mockgun

import pyshotgrid as pysg
import pyshotgrid.sg_default_entities as sde

from .testbase import BaseShotGridTest


class TestSGSite(BaseShotGridTest):
    """Tests for `pyshotgrid` package."""

    def test_people(self):
        sg_site = pysg.SGSite(self.sg)

        result = sg_site.people()

        self.assertEqual(3, len(result))
        self.assertTrue(isinstance(result[0], sde.SGHumanUser))

    def test_pipeline_configuration(self):
        sg_site = pysg.SGSite(self.sg)

        result = sg_site.pipeline_configuration()

        self.assertTrue(isinstance(result, sde.SGEntity))

    def test_pipeline_configuration__by_name(self):
        sg_site = pysg.SGSite(self.sg)

        result = sg_site.pipeline_configuration(name_or_id="Primary")

        self.assertTrue(isinstance(result, sde.SGEntity))
        self.assertEqual("Primary", result["code"].get())

    def test_pipeline_configuration__by_id(self):
        sg_site = pysg.SGSite(self.sg)

        result = sg_site.pipeline_configuration(name_or_id=1)

        self.assertTrue(isinstance(result, sde.SGEntity))
        self.assertEqual(1, result.id)

    def test_pipeline_configuration__by_project(self):
        sg_project = pysg.new_entity(self.sg, "Project", 1)
        sg_site = pysg.SGSite(self.sg)

        result = sg_site.pipeline_configuration(project=sg_project)

        self.assertTrue(isinstance(result, sde.SGEntity))
        self.assertEqual(sg_project, result["project"].get())

    def test_entity_field_schemas(self):
        sg_site = pysg.SGSite(self.sg)

        result = sg_site.entity_field_schemas()

        self.assertTrue(isinstance(result["Project"]["code"], pysg.FieldSchema))

    def test_comparison(self):
        sg_site_a = pysg.SGSite(self.sg)
        sg_site_b = pysg.SGSite(self.sg)

        result = sg_site_a == sg_site_b

        self.assertTrue(result)

    def test_comparison__inequality(self):
        sg_site_a = pysg.SGSite(self.sg)
        other_sg = mockgun.Shotgun(
            base_url="https://other.shotgunstudio.com",
            script_name="Unittest User",
            api_key="$ome_password",
        )
        sg_site_b = pysg.SGSite(other_sg)

        result = sg_site_a != sg_site_b

        self.assertTrue(result)

    def test_comparison__wrong_type(self):
        sg_site_a = pysg.SGSite(self.sg)

        result = sg_site_a != 1

        self.assertTrue(result)



import os
import sys
import unittest

from shotgun_api3.lib import mockgun


class BaseShotGridTest(unittest.TestCase):
    """
    Base Test Class to setup Mockgun.
    """

    sg = None

    @classmethod
    def setUpClass(cls):
        resources_dir = os.path.join(os.path.dirname(__file__), "resources")

        mockgun_schema_dir = os.path.join(
            resources_dir, "mockgun_schemas", "py{}".format(sys.version_info.major)
        )
        mockgun.Shotgun.set_schema_paths(
            os.path.join(mockgun_schema_dir, "schema.db"),
            os.path.join(mockgun_schema_dir, "entity_schema.db"),
        )
        cls.sg = mockgun.Shotgun(
            base_url="https://test.shotgunstudio.com",
            script_name="Unittest User",
            api_key="$ome_password",
        )

        cls.add_default_entities()

    @classmethod
    def add_default_entities(cls):
        # HumanUsers
        person_a = cls.sg.create(
            "HumanUser",
            {
                "firstname": "Alice",
                "lastname": "Alpha",
                "name": "Alice Alpha",
                "login": "alice.alpha",
                "email": "alice@company.com",
            },
        )
        person_b = cls.sg.create(
            "HumanUser",
            {
                "firstname": "Bob",
                "lastname": "Beta",
                "name": "Bob Beta",
                "login": "bob.beta",
                "email": "bob@company.com",
            },
        )
        person_c = cls.sg.create(
            "HumanUser",
            {
                "firstname": "Carol",
                "lastname": "Clown",
                "name": "Carol Clown",
                "login": "carol.clown",
                "email": "carol@company.com",
            },
        )

        # LocalStorages
        local_storage = cls.sg.create(
            "LocalStorage",
            {
                "code": "primary",
                "mac_path": "/Volumes/projects",
                "windows_path": "P:\\",
                "linux_path": "/mnt/projects",
            },
        )

        # Pipeline Steps
        pipeline_step_cmp = cls.sg.create(
            "Step",
            {
                "code": "Compositing",
                "short_name": "CMP",
            },
        )
        pipeline_step_lgt = cls.sg.create(
            "Step",
            {
                "code": "Lighting",
                "short_name": "LGT",
            },
        )

        # Projects
        project_a = cls.sg.create(
            "Project",
            {
                "name": "Test Project A",
                "tank_name": "tpa",
                "users": [person_a, person_b, person_c],
            },
        )
        project_b = cls.sg.create(
            "Project",
            {
                "name": "Test Project B",
                "tank_name": "tpb",
                "users": [person_a, person_b, person_c],
            },
        )

        # Assets
        cls.sg.create("Asset", {"code": "Tree", "project": project_a})
        cls.sg.create("Asset", {"code": "CarA", "project": project_a})
        cls.sg.create("Asset", {"code": "CarB", "project": project_a})

        # Sequences
        sq111 = cls.sg.create("Sequence", {"code": "sq111", "project": project_a})
        sq222 = cls.sg.create("Sequence", {"code": "sq222", "project": project_a})

        # Shots
        shot_a = cls.sg.create(
            "Shot", {"code": "sq111_sh1111", "project": project_a, "sg_sequence": sq111}
        )
        cls.sg.create(
            "Shot", {"code": "sq111_sh2222", "project": project_a, "sg_sequence": sq111}
        )
        cls.sg.create(
            "Shot", {"code": "sq222_sh3333", "project": project_a, "sg_sequence": sq222}
        )
        cls.sg.create(
            "Shot", {"code": "sq222_sh4444", "project": project_a, "sg_sequence": sq222}
        )

        # Tasks
        task_a = cls.sg.create(
            "Task",
            {
                "content": "comp",
                "project": project_a,
                "entity": shot_a,
                "step": pipeline_step_cmp,
                "task_assignees": [person_a],
            },
        )
        task_b = cls.sg.create(
            "Task",
            {
                "content": "lighting",
                "project": project_a,
                "entity": shot_a,
                "step": pipeline_step_lgt,
                "task_assignees": [person_b],
            },
        )

        # PublishedFileTypes
        pub_type_alembic = cls.sg.create("PublishedFileType", {"code": "Alembic Cache"})
        pub_type_render = cls.sg.create("PublishedFileType", {"code": "Rendered Image"})

        # PublishedFiles
        for i in range(1, 6):
            cls.sg.create(
                "PublishedFile",
                {
                    "code": "sh1111_city_v{:03d}.%04d.exr".format(i),
                    "name": "sh1111_city",
                    "version_number": i,
                    "published_file_type": pub_type_render,
                    "project": project_a,
                    "task": task_a,
                    "entity": shot_a,
                    "path_cache": "tp/sequences/sq111/sh1111_city_v{:03d}.%04d.exr".format(
                        i
                    ),
                    "path_cache_storage": local_storage,
                },
            )
        for i in range(1, 6):
            cls.sg.create(
                "PublishedFile",
                {
                    "code": "sh1111_city_v{:03d}.abc".format(i),
                    "name": "sh1111_city",
                    "version_number": i,
                    "published_file_type": pub_type_alembic,
                    "project": project_a,
                    "task": task_b,
                    "entity": shot_a,
                    "path_cache": "tp/sequences/sq111/sh1111_city_v{:03d}.abc".format(
                        i
                    ),
                    "path_cache_storage": local_storage,
                },
            )

        # Playlists
        cls.sg.create("Playlist", {"code": "Playlist A", "project": project_a})
        cls.sg.create("Playlist", {"code": "Playlist B", "project": project_b})
        cls.sg.create("Playlist", {"code": "Playlist C"})

        # PipelineConfigurations
        cls.sg.create("PipelineConfiguration", {"code": "Primary"})
        cls.sg.create(
            "PipelineConfiguration", {"code": "Primary", "project": project_a}
        )
        cls.sg.create(
            "PipelineConfiguration", {"code": "Develop", "project": project_a}
        )

        # Notes
        cls.sg.create("Note", {"subject": "Test note"})



"""Tests for `pyshotgrid.core` module."""

import pyshotgrid as pysg

from .testbase import BaseShotGridTest


class TestCoreConvertFunctions(BaseShotGridTest):
    def test_convert_filters_to_dict(self):
        person = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=2)

        result = pysg.core.convert_filters_to_dict([["user", "is", person]])

        self.assertEqual([["user", "is", {"type": "HumanUser", "id": 2}]], result)

    def test_convert_filters_to_dict__list_of_entities(self):
        person_a = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=1)
        person_b = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=2)

        result = pysg.core.convert_filters_to_dict(
            [["user", "in", [person_a, person_b]]]
        )

        self.assertEqual(
            [
                [
                    "user",
                    "in",
                    [{"type": "HumanUser", "id": 1}, {"type": "HumanUser", "id": 2}],
                ]
            ],
            result,
        )

    def test_convert_filters_to_dict__list_of_non_entity_values(self):
        result = pysg.core.convert_filters_to_dict([["priority", "in", [10, 20, 30]]])

        self.assertEqual([["priority", "in", [10, 20, 30]]], result)

    def test_convert_value_to_pysg__no_entity_value(self):
        example_value = "some other value than an entity"

        result = pysg.core.convert_value_to_pysg(self.sg, example_value)

        self.assertEqual(example_value, result)

    def test_convert_value_to_pysg__single_entity(self):
        person_a = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=1)

        result = pysg.core.convert_value_to_pysg(
            self.sg, {"type": "HumanUser", "id": 1}
        )

        self.assertEqual(person_a, result)

    def test_convert_value_to_pysg__list_of_entities(self):
        person_a = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=1)
        person_b = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=2)

        result = pysg.core.convert_value_to_pysg(
            self.sg, [{"type": "HumanUser", "id": 1}, {"type": "HumanUser", "id": 2}]
        )

        self.assertEqual([person_a, person_b], result)

    def test_convert_value_to_dict__no_entity_value(self):
        example_value = "some other value than an entity"

        result = pysg.core.convert_value_to_dict(example_value)

        self.assertEqual(example_value, result)

    def test_convert_value_to_dict__single_entity(self):
        person_a = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=1)

        result = pysg.core.convert_value_to_dict(person_a)

        self.assertEqual({"type": "HumanUser", "id": 1}, result)

    def test_convert_value_to_dict__list_of_entities(self):
        person_a = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=1)
        person_b = pysg.SGEntity(self.sg, entity_type="HumanUser", entity_id=2)

        result = pysg.core.convert_value_to_dict([person_a, person_b])

        self.assertEqual(
            [{"type": "HumanUser", "id": 1}, {"type": "HumanUser", "id": 2}], result
        )


class TestCoreNewFunctions(BaseShotGridTest):
    def setUp(self):
        pysg.core.__ENTITY_PLUGINS = {}

    def test_new_entity(self):
        sg_entity_a = pysg.new_entity(self.sg, {"type": "Project", "id": 1})
        sg_entity_b = pysg.new_entity(self.sg, "Project", 1)
        sg_entity_c = pysg.new_entity(self.sg, entity_type="Project", entity_id=1)

        self.assertTrue(sg_entity_a == sg_entity_b == sg_entity_c)

    def test_new_entity__raises_error_on_wrong_inputs(self):
        with self.assertRaises(ValueError):
            pysg.new_entity(self.sg, 123)

    def test_new_site(self):
        sg_site_a = pysg.new_site(self.sg)

        self.assertTrue(isinstance(sg_site_a, pysg.SGSite))


class SGNote(pysg.SGEntity):
    """
    Test SGEntity class.
    """


class CustomSGProject(pysg.sg_default_entities.SGProject):
    """
    Test class from default SGEntity.
    """


class CustomSGSite(pysg.SGSite):
    """
    Test SGSite class.
    """


class TestCorePluginFunctions(BaseShotGridTest):
    def test_register_pysg_class__add_custom_sg_class(self):
        pysg.register_pysg_class("Note", SGNote)
        sg_entity_a = pysg.new_entity(self.sg, {"type": "Note", "id": 1})

        result = sg_entity_a.__class__

        self.assertTrue(SGNote == result)

    def test_register_pysg_class__overwrite_default_sg_class(self):
        pysg.register_pysg_class("Project", CustomSGProject)
        sg_entity_a = pysg.new_entity(self.sg, {"type": "Project", "id": 1})

        result = sg_entity_a.__class__

        self.assertTrue(CustomSGProject == result)

    def test_register_pysg_class__errors_on_invalid_class(self):
        with self.assertRaises(TypeError):
            pysg.register_pysg_class("Project", str)  # type: ignore

    def test_register_sg_site_class__add_custom_class(self):
        pysg.register_sg_site_class(CustomSGSite)
        sg_entity_a = pysg.new_site(self.sg)

        result = sg_entity_a.__class__

        self.assertTrue(CustomSGSite == result)

    def test_register_sg_site_class__errors_on_invalid_class(self):
        with self.assertRaises(TypeError):
            pysg.register_sg_site_class(str)  # type: ignore



"""
This module collects all default pyshotgrid custom entities.
"""
import fnmatch
import typing
from typing import Any, Dict, List, Optional, Type, Union  # noqa: F401

if typing.TYPE_CHECKING:
    import shotgun_api3  # noqa: F401

    from .core import Field  # noqa: F401

from .core import SGEntity, new_entity


class SGProject(SGEntity):
    """
    An instance of this class represents a single Project entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def shots(self, glob_pattern=None):
        # type: (Optional[str]) -> List[SGEntity]
        """
        :param glob_pattern: A glob to match the shots to return. For example
                             `TEST_01_*` would return all shots that start with `TEST_01_`.
        :return: All the shots from this project.
        """
        sg_shots = self.sg.find("Shot", [["project", "is", self.to_dict()]], ["code"])
        if glob_pattern is not None:
            return [
                new_entity(self._sg, sg_shot)
                for sg_shot in sg_shots
                if fnmatch.fnmatchcase(sg_shot["code"], glob_pattern)
            ]
        else:
            return [new_entity(self._sg, sg_shot) for sg_shot in sg_shots]

    def assets(self, glob_pattern=None):
        # type: (Optional[str]) -> List[SGEntity]
        """
        :param glob_pattern: A glob to match the assets to return. For example
                            `TEST_*` would return all assets that start with `TEST_`.
        :return: All the assets from this project.
        """
        sg_assets = self.sg.find("Asset", [["project", "is", self.to_dict()]], ["code"])
        if glob_pattern is not None:
            return [
                new_entity(self._sg, sg_asset)
                for sg_asset in sg_assets
                if fnmatch.fnmatchcase(sg_asset["code"], glob_pattern)
            ]
        else:
            return [new_entity(self._sg, sg_asset) for sg_asset in sg_assets]

    def publishes(
        self,
        pub_types=None,  # type: Optional[Union[str,List[str]]]
        latest=False,  # type: bool
        additional_sg_filter=None,  # type: Optional[List[Any]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param pub_types: The names of the Publish File Types to return.
        :param latest: Whether to get the "latest" publishes or not. This uses the
          same logic as the tk-multi-loader2 app which is as follows:

            - group all publishes with the same "name" field together
            - from these get the publishes with the highest "version_number" field
            - if there are publishes with the same "name" and "version_number" the
              newest one wins.
        :param additional_sg_filter:
        :return: All published files from this project.
        """
        return self._publishes(
            base_filter=[["project", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )

    def people(self, additional_sg_filter=None):
        # type: (Optional[List[Any]]) -> List[SGEntity]
        """
        :param additional_sg_filter:
        :return: All HumanUsers assigned to this project.
        """
        sg_filter = [["projects", "contains", self.to_dict()]]
        if additional_sg_filter is not None:
            sg_filter += additional_sg_filter

        return [
            new_entity(self._sg, sg_user)
            for sg_user in self._sg.find("HumanUser", sg_filter)
        ]

    def playlists(self):
        # type: () -> List[SGEntity]
        """
        :return: All playlists attached to this project.
        """
        return [
            new_entity(self._sg, sg_playlist)
            for sg_playlist in self._sg.find(
                "Playlist", [["project", "is", self.to_dict()]]
            )
        ]


class SGShot(SGEntity):
    """
    An instance of this class represents a single Shot entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def publishes(
        self,
        pub_types=None,  # type: Optional[Union[str,List[str]]]
        latest=False,  # type: bool
        additional_sg_filter=None,  # type: Optional[List[Any]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param pub_types: The names of the Publish File Types to return.
        :param latest: Whether to get the "latest" publishes or not. This uses the
                       same logic as the tk-multi-loader2 app which is as follows:

                         - group all publishes with the same "name" field together
                         - from these get the publishes with the highest "version_number" field
                         - if there are publishes with the same "name" and "version_number" the
                           newest one wins.
        :param additional_sg_filter:
        :return: All published files from this shot.
        """
        return self._publishes(
            base_filter=[["entity", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )

    def tasks(
        self,
        names=None,  # type: Optional[List[str]]
        pipeline_step=None,  # type: Optional[Union[str,Dict[str,Any],SGEntity]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param names: The names of Tasks to return.
        :param pipeline_step: Name, short name or entity object
                              or the Pipeline Step to filter by.
        :returns: A list of Tasks
        """
        sg_filter = [
            ["entity", "is", self.to_dict()]
        ]  # type: List[Union[List[Any],Dict[str,Any]]]

        if names is not None:
            if len(names) == 1:
                names_filter = [
                    "code",
                    "is",
                    names[0],
                ]  # type: Union[List[Any],Dict[str,Any]]
            else:
                names_filter = {"filter_operator": "any", "filters": []}
                for name in names:
                    names_filter["filters"].append(["code", "is", name])
            sg_filter.append(names_filter)

        if pipeline_step is not None:
            if isinstance(pipeline_step, dict):
                sg_filter.append(["step", "is", pipeline_step])
            elif isinstance(pipeline_step, SGEntity):
                sg_filter.append(["step", "is", pipeline_step.to_dict()])
            else:
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["step.Step.code", "is", pipeline_step],
                            ["step.Step.short_name", "is", pipeline_step],
                        ],
                    }
                )

        return [
            new_entity(self._sg, sg_task)
            for sg_task in self._sg.find("Task", sg_filter)
        ]


class SGAsset(SGEntity):
    """
    An instance of this class represents a single Asset entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def publishes(
        self,
        pub_types=None,  # type: Optional[Union[str,List[str]]]
        latest=False,  # type: bool
        additional_sg_filter=None,  # type: Optional[List[Any]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param pub_types: The names of the Publish File Types to return.
        :param latest: Whether to get the "latest" publishes or not. This uses the
                       same logic as the tk-multi-loader2 app which is as follows:

                         - group all publishes with the same "name" field together
                         - from these get the publishes with the highest "version_number" field
                         - if there are publishes with the same "name" and "version_number" the
                           newest one wins.
        :param additional_sg_filter:
        :return: All published files from this asset.
        """
        return self._publishes(
            base_filter=[["entity", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )

    def tasks(
        self,
        names=None,  # type: Optional[List[str]]
        pipeline_step=None,  # type: Optional[Union[str,Dict[str,Any],SGEntity]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param names: The names of Tasks to return.
        :param pipeline_step: Name, short name or entity object
                              or the Pipeline Step to filter by.
        :returns: A list of Tasks.
        """
        sg_filter = [
            ["entity", "is", self.to_dict()]
        ]  # type: List[Union[List[Any],Dict[str,Any]]]

        if names is not None:
            if len(names) == 1:
                names_filter = [
                    "code",
                    "is",
                    names[0],
                ]  # type: Union[List[Any],Dict[str,Any]]
            else:
                names_filter = {"filter_operator": "any", "filters": list()}
                for name in names:
                    names_filter["filters"].append(["code", "is", name])
            sg_filter.append(names_filter)

        if pipeline_step is not None:
            if isinstance(pipeline_step, dict):
                sg_filter.append(["step", "is", pipeline_step])
            elif isinstance(pipeline_step, SGEntity):
                sg_filter.append(["step", "is", pipeline_step.to_dict()])
            else:
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["step.Step.code", "is", pipeline_step],
                            ["step.Step.short_name", "is", pipeline_step],
                        ],
                    }
                )

        return [
            new_entity(self._sg, sg_task)
            for sg_task in self._sg.find("Task", sg_filter)
        ]


class SGTask(SGEntity):
    """
    An instance of this class represents a single Task entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    @property
    def name(self):
        # type: () -> Field
        """
        :return: The field that represents the name of the Task.
        """
        return self["content"]

    def publishes(
        self,
        pub_types=None,  # type: Optional[Union[str,List[str]]]
        latest=False,  # type: bool
        additional_sg_filter=None,  # type: Optional[List[Any]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param pub_types: The names of the Publish File Types to return.
        :param latest: Whether to get the "latest" publishes or not. This uses the
                       same logic as the tk-multi-loader2 app which is as follows:

                        - group all publishes with the same "name" field together
                        - from these get the publishes with the highest "version_number" field
                        - if there are publishes with the same "name" and "version_number" the
                          newest one wins.
        :param additional_sg_filter:
        :return: All published files from this shot.
        """
        return self._publishes(
            base_filter=[["task", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )


class SGPublishedFile(SGEntity):
    """
    An instance of this class represents a single PublishedFile entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def is_latest(self):
        # type: () -> bool
        """
        :return: Whether this published file is the latest of its kind.
        """
        return self.get_latest_publish() == self

    def get_latest_publish(self):
        # type: () -> SGEntity
        """
        :return: The latest published file of its kind (which might be this same entity).
        """
        return self.get_all_publishes()[-1]

    def get_next_publishes(self):
        # type: () -> List[SGEntity]
        """
        :return: The next publishes after this publish.
        """
        all_publishes = self.get_all_publishes()
        index = all_publishes.index(self)
        return all_publishes[index + 1 :]

    def get_previous_publishes(self):
        # type: () -> List[SGEntity]
        """
        :return: The previous publishes before this publish.
        """
        all_publishes = self.get_all_publishes()
        index = all_publishes.index(self)
        return all_publishes[:index]

    def get_all_publishes(self):
        # type: () -> List[SGEntity]
        """
        :return: A list of all the published file versions from lowest to highest version number.
        """
        this_publish = self.get(
            ["entity", "published_file_type", "name"], raw_values=True
        )
        sg_publishes = self.sg.find(
            "PublishedFile",
            [
                ["entity", "is", this_publish["entity"]],
                ["published_file_type", "is", this_publish["published_file_type"]],
                ["name", "is", this_publish["name"]],
            ],
            ["name", "version_number", "created_at"],
        )

        # sort them by date and than by version_number which sorts the latest publish to the
        # last position.
        sg_publishes.sort(key=lambda pub: (pub["created_at"], pub["version_number"]))

        return [new_entity(self._sg, sg_publish) for sg_publish in sg_publishes]


class SGVersion(SGEntity):
    """
    An instance of this class represents a single Version entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """


class SGPlaylist(SGEntity):
    """
    An instance of this class represents a single Playlist entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    @property
    def media_url(self):
        # type: () -> str
        """
        :return: The Media center URL for this playlist.
        :raises:
            :RuntimeError: When this playlist is not attached to a project.
        """
        sg_project = self["project"].get()
        if sg_project is None:
            raise RuntimeError(
                'Cannot get media URL for playlist "{}"'
                ", because it is not attached to a project.".format(self.id)
            )
        # Example URL:
        # https://example.shotgunstudio.com/page/media_center?type=Playlist&id=123&project_id=456
        return "{}/page/media_center?type={}&id={}&project_id={}".format(
            self.sg.base_url, self._type, self._id, sg_project["id"].get()
        )


class SGHumanUser(SGEntity):
    """
    An instance of this class represents a single HumanUser entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def tasks(
        self,
        names=None,  # type: Optional[List[str]]
        project=None,  # type: Optional[Union[str,Dict[str,Any],SGEntity]]
        pipeline_step=None,  # type: Optional[Union[str,Dict[str,Any],SGEntity]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param list[str]|None names: The names of Tasks to return.
        :param str|dict|SGEntity|None project: Name, tank_name or entity object
                                               or the Project to filter by.
        :param str|dict|SGEntity|None pipeline_step: Name, short name or entity object
                                                     or the Pipeline Step to filter by.
        :returns: A list of Tasks
        :rtype: list[SGTask]
        """
        sg_filter = [
            {
                "filter_operator": "any",
                "filters": [
                    ["task_assignees", "contains", self.to_dict()],
                    ["task_assignees.Group.users", "contains", self.to_dict()],
                ],
            }
        ]  # type: List[Union[List[Any],Dict[str,Any]]]

        if names is not None:
            if len(names) == 1:
                names_filter = [
                    "code",
                    "is",
                    names[0],
                ]  # type: Union[List[Any],Dict[str,Any]]
            else:
                names_filter = {"filter_operator": "any", "filters": []}
                for name in names:
                    names_filter["filters"].append(["code", "is", name])
            sg_filter.append(names_filter)

        if project is not None:
            if isinstance(project, dict):
                sg_filter.append(["project", "is", project])
            elif isinstance(project, SGEntity):
                sg_filter.append(["project", "is", project.to_dict()])
            elif isinstance(project, str):
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["project.Project.code", "is", project],
                            ["project.Project.tank_name", "is", project],
                        ],
                    }
                )
            else:
                raise TypeError(
                    'The "project" parameter needs to be one of '
                    "type str, dict, SGEntity or None."
                )

        if pipeline_step is not None:
            if isinstance(pipeline_step, dict):
                sg_filter.append(["step", "is", pipeline_step])
            elif isinstance(pipeline_step, SGEntity):
                sg_filter.append(["step", "is", pipeline_step.to_dict()])
            elif isinstance(pipeline_step, str):
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["step.Step.code", "is", pipeline_step],
                            ["step.Step.short_name", "is", pipeline_step],
                        ],
                    }
                )
            else:
                raise TypeError(
                    'The "pipeline_step" parameter needs to be one of '
                    "type str, dict, SGEntity or None."
                )

        return [
            new_entity(self._sg, sg_task)
            for sg_task in self._sg.find("Task", sg_filter)
        ]

    def publishes(
        self,
        pub_types=None,  # type: Optional[Union[str,List[str]]]
        latest=False,  # type: bool
        additional_sg_filter=None,  # type: Optional[List[Any]]
    ):
        # type: (...) -> List[SGEntity]
        """
        :param pub_types: The names of the Publish File Types to return.
        :param latest: Whether to get the "latest" publishes or not. This uses the
                       same logic as the tk-multi-loader2 app which is as follows:

                        - group all publishes with the same "name" field together
                        - from these get the publishes with the highest "version_number" field
                        - if there are publishes with the same "name" and "version_number" the
                          newest one wins.
        :param additional_sg_filter:
        :return: All published files from this shot.
        """
        return self._publishes(
            base_filter=[["created_by", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )



#!/usr/bin/env python

"""Tests for `pyshotgrid` package."""
import sys

if sys.version_info.major == 2:
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    import mock
else:
    from unittest import mock

from shotgun_api3.lib import mockgun

import pyshotgrid as pysg

from .testbase import BaseShotGridTest


class TestSGEntity(BaseShotGridTest):
    def test_string_representation(self):
        sg_entity = pysg.SGEntity(self.sg, "Project", 1)

        self.assertEqual(
            "SGEntity - Type: Project - ID: 1 - "
            "URL: https://test.shotgunstudio.com/detail/Project/1",
            str(sg_entity),
        )

    def test_query_field_dict_notation(self):
        sg_entity = pysg.SGEntity(self.sg, "Project", 1)

        self.assertEqual("tpa", sg_entity["tank_name"].get())

    def test_get(self):
        sg_entity = pysg.SGEntity(self.sg, "Project", 1)

        self.assertEqual(
            {"name": "Test Project A", "tank_name": "tpa"},
            sg_entity.get(["name", "tank_name"]),
        )

    def test_name(self):
        sg_project = pysg.new_entity(self.sg, "Project", 1)
        sg_task = pysg.new_entity(self.sg, "Task", 1)
        sg_asset = pysg.new_entity(self.sg, "Asset", 1)

        result_project_name_field = sg_project.name
        result_task_name_field = sg_task.name
        result_asset_name_field = sg_asset.name

        self.assertEqual("name", result_project_name_field.name)
        self.assertEqual("content", result_task_name_field.name)
        self.assertEqual("code", result_asset_name_field.name)

    def test_name__errors_when_no_name_field_present(self):
        sg_entity = pysg.new_entity(self.sg, "Note", 1)

        with self.assertRaises(RuntimeError):
            _ = sg_entity.name

    # FIXME Mockgun.update is missing the "multi_entity_update_modes" parameter. We need to patch it
    # FIXME to make this test work.
    # def test_set(self):
    #     sg_entity = pysg.SGEntity(self.sg, 'Project', 1)
    #     def patched_update(entity_type, entity_id, data, multi_entity_update_modes=None):
    #         self.sg.update(entity_type, entity_id, data)
    #
    #     with mock.patch.object(mockgun.Shotgun, "update", new_callable=patched_update):
    #         sg_entity.set({'code': 'Test Name', 'tank_name': 'tn'})
    #
    #     self.assertEqual({'code': 'Test Name', 'tank_name': 'tn'},
    #                      sg_entity.get(['code', 'tank_name']))
    #     # Cleanup
    #     sg_entity.set({'code': 'Test Project', 'tank_name': 'tp'})

    def test_update_field_dict_notation(self):
        sg_entity = pysg.SGEntity(self.sg, "Project", 1)
        sg_entity["code"].set("bar-baz")

        self.assertEqual("bar-baz", sg_entity["code"].get())

        # cleanup
        sg_entity["code"].set("Test project")

    def test_convert_to_dict(self):
        # published files to query publish form paths.
        sg_publish = self.sg.create("PublishedFile", {"code": "delete_me"})
        sg_entity = pysg.SGEntity(self.sg, sg_publish["type"], sg_publish["id"])

        sg_entity.delete()

        self.assertEqual(
            [], self.sg.find(sg_publish["type"], [["id", "is", sg_publish["id"]]])
        )

    def test_delete(self):
        sg_entity = pysg.SGEntity(self.sg, "LocalStorage", 1)

        self.assertEqual({"id": 1, "type": "LocalStorage"}, sg_entity.to_dict())

    def test_site(self):
        sg_entity = pysg.SGEntity(self.sg, "LocalStorage", 1)

        self.assertIsInstance(sg_entity.site, pysg.SGSite)

    def test_iter_all_field_values(self):
        sg_entity = pysg.SGEntity(self.sg, "LocalStorage", 1)

        # Mock Mockgun.schema_field_read - the "project_entity" arg is missing in Mockgun.
        with mock.patch.object(
            mockgun.Shotgun,
            "schema_field_read",
            return_value={
                "linux_path": {"visible": {"value": True}},
                "type": {"visible": {"value": True}},
                "id": {"visible": {"value": True}},
                "description": {"visible": {"value": True}},
                "cached_display_name": {"visible": {"value": True}},
                "mac_path": {"visible": {"value": True}},
                "created_at": {"visible": {"value": True}},
                "created_by": {"visible": {"value": True}},
                "windows_path": {"visible": {"value": True}},
                "uuid": {"visible": {"value": True}},
                "code": {"visible": {"value": True}},
                "updated_at": {"visible": {"value": True}},
                "updated_by": {"visible": {"value": True}},
            },
        ):
            result = set(sg_entity.all_field_values().items())

        self.assertEqual(
            {
                ("linux_path", "/mnt/projects"),
                ("type", "LocalStorage"),
                ("id", 1),
                ("description", None),
                ("cached_display_name", None),
                ("mac_path", "/Volumes/projects"),
                ("created_by", None),
                ("updated_by", None),
                ("created_at", None),
                ("windows_path", "P:\\"),
                ("uuid", None),
                ("code", "primary"),
                ("updated_at", None),
            },
            result,
        )

    def test_fields(self):
        sg_entity = pysg.SGEntity(self.sg, "LocalStorage", 1)

        # Mock Mockgun.schema_field_read - the "project_entity" arg is missing in Mockgun.
        with mock.patch.object(
            mockgun.Shotgun,
            "schema_field_read",
            return_value={
                "linux_path": {"visible": {"value": True}},
                "type": {"visible": {"value": True}},
                "id": {"visible": {"value": True}},
                "description": {"visible": {"value": True}},
                "cached_display_name": {"visible": {"value": True}},
                "mac_path": {"visible": {"value": True}},
                "created_at": {"visible": {"value": True}},
                "created_by": {"visible": {"value": True}},
                "windows_path": {"visible": {"value": True}},
                "uuid": {"visible": {"value": True}},
                "code": {"visible": {"value": True}},
                "updated_at": {"visible": {"value": True}},
                "updated_by": {"visible": {"value": True}},
            },
        ):
            result = sg_entity.fields()

        self.assertEqual(
            {
                "linux_path",
                "type",
                "id",
                "description",
                "cached_display_name",
                "mac_path",
                "created_by",
                "updated_by",
                "created_at",
                "windows_path",
                "uuid",
                "code",
                "updated_at",
            },
            {field.name for field in result},
        )
        self.assertTrue(all([isinstance(field, pysg.Field) for field in result]))

    def test_shotgun_url(self):
        sg_entity = pysg.SGEntity(self.sg, "Project", 1)

        self.assertEqual(
            "https://test.shotgunstudio.com/detail/Project/1", sg_entity.url
        )

    def test_entity_display_name(self):
        sg_entity = pysg.SGEntity(self.sg, "CustomEntity01", 1)

        self.assertEqual("Sprint", sg_entity.entity_display_name)

    def test_schema(self):
        sg_entity = pysg.SGEntity(self.sg, "Project", 1)

        self.assertEqual(
            {
                "name": {"editable": False, "value": "Project"},
                "visible": {"editable": False, "value": True},
            },
            sg_entity.schema(),
        )

    def test_field_schemas(self):
        sg_entity = pysg.SGEntity(self.sg, "Project", 1)

        result = sg_entity.field_schemas()

        self.assertEqual("Tank Name", result["tank_name"].display_name)
        self.assertEqual(67, len(result))

    def test_compare_entities(self):
        sg_entity_a = pysg.SGEntity(self.sg, "Project", 1)
        sg_entity_b = pysg.sg_default_entities.SGProject(self.sg, "Project", 1)

        result = sg_entity_a == sg_entity_b

        self.assertTrue(result)

    def test_compare_entities__ids_dont_match(self):
        sg_entity_a = pysg.SGEntity(self.sg, "Project", 1)
        sg_entity_b = pysg.SGEntity(self.sg, "Project", 2)

        result = sg_entity_a != sg_entity_b

        self.assertTrue(result)

    def test_compare_entities__entity_types_dont_match(self):
        sg_entity_a = pysg.SGEntity(self.sg, "Project", 1)
        sg_entity_b = pysg.SGEntity(self.sg, "Version", 1)

        result = sg_entity_a != sg_entity_b

        self.assertTrue(result)

    def test_compare_entities__sg_instances_dont_match(self):
        sg_entity_a = pysg.SGEntity(self.sg, "Project", 1)
        other_sg = mockgun.Shotgun(
            base_url="https://other.shotgunstudio.com",
            script_name="Unittest User",
            api_key="$ome_password",
        )
        sg_entity_b = pysg.SGEntity(other_sg, "Project", 1)

        result = sg_entity_a != sg_entity_b

        self.assertTrue(result)

    def test_compare_entities__not_a_sg_entity(self):
        sg_entity_a = pysg.SGEntity(self.sg, "Project", 1)

        result = sg_entity_a != 1

        self.assertTrue(result)
